/*
 * bl - backlight control
 *
 *
 *** This bit is from the original bl ***
 *
 * Copyright 2000 Compaq Computer Corporation.
 *
 * Use consistent with the GNU GPL is permitted,
 * provided that this copyright notice is
 * preserved in its entirety in all copies and derived works.
 *
 * COMPAQ COMPUTER CORPORATION MAKES NO WARRANTIES, EXPRESSED OR IMPLIED,
 * AS TO THE USEFULNESS OR CORRECTNESS OF THIS CODE OR ITS
 * FITNESS FOR ANY PARTICULAR PURPOSE.
 *
 * Author: Charles Flynn.
 * 
 * Updated by Andrew Christian
 *            October, 2001
 *
 *** This bit is from minilite ***
 * 
 * Copyright 2002, 2003 Phil Blundell
 *           2004, 2005 Florian Boor <florian@kernelconcepts.de>
 * Contributions by: Arjan Schrijver, Koen Kooi and Nils Faerber
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ***
 *
 * Hacked together by Simon Pickering <S.G.Pickering@bath.ac.uk>
 *
 ***
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <libintl.h>
#include <dirent.h>

#include <linux/ioctl.h>
#include <sys/ioctl.h>

typedef enum
{
	P_NONE,
	P_IPAQ,
	P_ZAURUS,
	P_CORGI,
	P_INTEGRAL,
	P_SIMPAD,
	P_SIMPAD_NEW,
    P_GENERIC,
	P_SYSCLASS_770,
	P_SYSCLASS
}t_platform;


int debug=0;

/* for iPAQ touchscreen */
#define IOC_H3600_TS_MAGIC  'f'
#define TS_GET_BACKLIGHT        _IOR(IOC_H3600_TS_MAGIC, 20, struct h3600_ts_backlight)
#define TS_SET_BACKLIGHT        _IOW(IOC_H3600_TS_MAGIC, 20, struct h3600_ts_backlight)

enum flite_pwr {
        FLITE_PWR_OFF = 0,
        FLITE_PWR_ON  = 1
};

struct h3600_ts_backlight {
        enum flite_pwr power;          /* 0 = off, 1 = on */
        unsigned char  brightness;     /* 0 - 255         */
};

#define TS_DEV "/dev/touchscreen/0raw"
#define H36XX 0
#define H38XX 1
#define H5XXX 2
#define H19XX 3
#define H22XX 4
char IpaqModel = -1;

/* for Integral 90200 */
#define PROC_LIGHT "/proc/hw90200/backlight"

/* for Sharp Corgi */
#define CORGI_FL "/proc/driver/fl/corgi-bl"

/* other Zauri */
#define ZAURUS_FL "/dev/fl"
#define FL_IOCTL_STEP_CONTRAST    100

/* Simpad, a little bit nasty - PWM reg. is set directly */
#define SIMPAD_BACKLIGHT_REG		"/proc/driver/mq200/registers/PWM_CONTROL"
#define SIMPAD_BACKLIGHT_MASK		0x00a10044

/* Simpad, new interface */
#define SIMPAD_BACKLIGHT_REG_NEW	"/proc/driver/mq200/backlight"

/* Generic backlight */
#define GENERIC_PROC_DRIVER "/proc/driver/backlight"

/* 2.6 style /sys/class/backlight */
#define SYSCLASS 	"/sys/class/backlight/"
#define SYS_STATE_ON  0
#define SYS_STATE_OFF 4
static char *SYS_BRIGHTNESS = NULL;
static char *SYS_MAXBRIGHTNESS = NULL;
static char *SYS_POWER = NULL;
/* wrong sysclass path on Nokia 770 */
#define SYSCLASS_770 	"/sys/devices/platform/omapfb/panel/"

#define _(_x)  gettext (_x)

t_platform platform = P_NONE;



void usage( void )
{
	fprintf(stderr,"Usage: bl MODE POWER BRIGHTNESS   Set using full triple [deprecated]\n");
	fprintf(stderr,"       bl                         Returns current frontlight/backlight\n");
	fprintf(stderr,"       bl VALUE                   Turns on and sets to VALUE (0 - 255)\n");
	fprintf(stderr,"       bl off                     Turns off frontlight/backlight\n");
	fprintf(stderr,"       bl on                      Turns on frontlight/backlight to last setting\n");
	fprintf(stderr,"Options:\n");
	fprintf(stderr,"       --help,-h                  Returns this message\n");
	fprintf(stderr,"\nPlease note: H3100 iPAQs have a screen backlight that can be either on or off (no level).\n");
	fprintf(stderr,"             H3600 and higher iPAQs have a screen frontlight that can be set from 0 to 255\n");
	fprintf(stderr,"             or turned off.\n");

	exit(0);
}




static char
*get_sysclass_bl(void)
{
	DIR *dp;
	char *dentry;
	struct dirent *d;

	if((dp = opendir(SYSCLASS)) == NULL) {
		fprintf(stderr, "unable to open %s", SYSCLASS);
		return NULL;
	}

	while((d = readdir(dp))) {
		if (!(strcmp(d->d_name, ".") == 0) &&
			!(strcmp(d->d_name, "..") == 0))
			asprintf(&dentry, "%s%s", SYSCLASS, d->d_name);
	}
	closedir(dp);
	return dentry;
}

static void
setup_sysclass(void)
{
	char *bl_dev;
	
	bl_dev = get_sysclass_bl();
	
	SYS_BRIGHTNESS = g_strdup_printf("%s/brightness", bl_dev);
	SYS_MAXBRIGHTNESS = g_strdup_printf("%s/max_brightness", bl_dev);
	SYS_POWER = g_strdup_printf("%s/power", bl_dev);
}

static void
setup_sysclass_770(void)
{
	SYS_BRIGHTNESS = g_strdup_printf("%s/backlight_level", SYSCLASS_770);
	SYS_MAXBRIGHTNESS = g_strdup_printf("%s/backlight_max", SYSCLASS_770);
	SYS_POWER = NULL;
}

static t_platform
detect_platform(void)
{
	if (!access(TS_DEV,R_OK))
		return P_IPAQ;
	if (!access(PROC_LIGHT,R_OK))
		return P_INTEGRAL;
	if (!access(CORGI_FL,R_OK))
		return P_CORGI;
	if (!access(ZAURUS_FL,R_OK))
		return P_ZAURUS;
	if (!access(SIMPAD_BACKLIGHT_REG_NEW,R_OK)) /* preserve order */
		return P_SIMPAD_NEW;
	if (!access(SIMPAD_BACKLIGHT_REG,R_OK))
		return P_SIMPAD;
	if (!access(GENERIC_PROC_DRIVER,R_OK))
		return P_GENERIC;
	if (!access(SYSCLASS_770, R_OK))
	{
		setup_sysclass_770();
		return P_SYSCLASS_770;
	}
	if (!access(SYSCLASS, R_OK))
	{
		setup_sysclass();
		return P_SYSCLASS;
	}
	
	return P_NONE;
}







int 
sysclass_set_level(int level)
{
  FILE *f_light;
  FILE *f_max;
  int val, maxlevel;
  
  if ( level > 255 ) level = 255;
  if ( level < 1 ) level = 1;

  f_max = fopen(SYS_MAXBRIGHTNESS, "r");
  if (f_max != NULL)
  {
    fscanf(f_max,"%d", &maxlevel);
    fclose(f_max);
  }
  val = ( level == 1 ) ? 1 : ( level * maxlevel ) / 255;
  if (val > maxlevel)
  {
    val = maxlevel;
  }
  f_light = fopen(SYS_BRIGHTNESS, "w");
  if (f_light != NULL)
  {
    fprintf(f_light,"%d\n", val);
  	fclose(f_light);
  }
  else
	  return -1;
  
  f_light = fopen(SYS_POWER, "w");
  if (f_light != NULL)
  {
    fprintf(f_light,"%d\n",  level ? SYS_STATE_ON : SYS_STATE_OFF);
  	fclose(f_light);
  }
  
  return level;
}


/* NEW */

int 
sysclass_set_power(int power_on)
{
  FILE *f_light;
  
  if ( power_on > 1 ) power_on = 1;
  if ( power_on < 0 ) power_on = 0;
  
 
  f_light = fopen(SYS_POWER, "w");
  if (f_light != NULL)
  {
    fprintf(f_light,"%d\n",  power_on ? SYS_STATE_ON : SYS_STATE_OFF);
  	fclose(f_light);
  }

 /*
	int fd = open( SYS_POWER, O_WRONLY|O_NONBLOCK );
	if ( fd )
	{
		char buf[10];
		buf[0] = on ? SYS_STATE_ON : SYS_STATE_OFF;
		buf[1] = '\0';
		res = ( write( fd, &buf[0], 2 ) == 0 );
		close( fd );
	}
  */

  return power_on;
}


int 
sysclass_get_level(void)
{
  FILE *f_light;
  FILE *f_max;
  float level, maxlevel, factor;
  
  f_max = fopen(SYS_MAXBRIGHTNESS, "r");
  if (f_max != NULL)
  {
    fscanf(f_max,"%f", &maxlevel);
    fclose(f_max);
  }
  factor = 255/maxlevel;
  f_light = fopen(SYS_BRIGHTNESS, "r");
  if (f_light != NULL)
  {
  	fscanf(f_light,"%f", &level);
  	fclose(f_light);
    level = level * factor;
    if (level > 255)
    {
      level = 255;
    }
	return (int)level;
  }
  return -1;
}  


/* NEW */

int 
sysclass_get_power(void)
{
  FILE *f_light;
  int power_on;

  f_light = fopen(SYS_POWER, "r");
  if (f_light != NULL)
  {
  	fscanf(f_light,"%f", &power_on);
  	fclose(f_light);
	power_on = (1 - power_on);
  }
  else power_on=1;
  
  
  return power_on;
}







int 
generic_set_level(int level)
{
  FILE *f_light;
  int adj_level;
  
  f_light = fopen(GENERIC_PROC_DRIVER,"w");
  if (f_light != NULL)
  {
	adj_level = level/255*32;
	fprintf(f_light,"%d\n", adj_level);
  	fclose(f_light);
	return level/32*255;
  }
  else
	  return -1;
}

int 
generic_set_power(int level)
{
	return 1;
}

int 
generic_get_level(void)
{
  FILE *f_light;
  int level;
  
  f_light = fopen(GENERIC_PROC_DRIVER,"r");
  if (f_light != NULL)
  {
  	fscanf(f_light,"%d", &level);
  	fclose(f_light);
	return level/32*255;
  }
  return -1;
}  

int 
generic_get_power(void)
{
	return 1;
}







int 
simpad_new_set_level(int level)
{
  FILE *f_light;
  
  if (level < 1) level = 1;
  f_light = fopen(SIMPAD_BACKLIGHT_REG_NEW,"w");
  if (f_light != NULL)
  {
    fprintf(f_light,"%d\n", level);
  	fclose(f_light);
	return level;
  }
  else
	  return -1;
}

int 
simpad_new_set_power(int level)
{
	return 1;
}

int 
simpad_new_get_level(void)
{
  FILE *f_light;
  int level;
  
  f_light = fopen(SIMPAD_BACKLIGHT_REG_NEW,"r");
  if (f_light != NULL)
  {
  	fscanf(f_light,"%d", &level);
  	fclose(f_light);
	return level;
  }
  return -1;
}  

int 
simpad_new_get_power(void)
{
	return 1;
}






int 
simpad_set_level(int level)
{
  int val;
  FILE *f_light;
  
  if (level < 1) level = 1;
  f_light = fopen(SIMPAD_BACKLIGHT_REG,"w");
  if (f_light != NULL)
  {
	val = 255 - level;
	val = val << 8;
	val = val + (int)SIMPAD_BACKLIGHT_MASK;
    fprintf(f_light,"0x%x\n", val);
  	fclose(f_light);
	return level;
  }
  else
	  return -1;
}

int 
simpad_set_power(int level)
{
	return 1;
}

int 
simpad_get_level(void)
{
  FILE *f_light;
  int level;
  
  f_light = fopen(SIMPAD_BACKLIGHT_REG,"r");
  if (f_light != NULL)
  {
  	fscanf(f_light,"0x%x", &level);
  	fclose(f_light);
	level = level - (int)SIMPAD_BACKLIGHT_MASK;
	level = level >> 8;
	return 255 - level;
  }
  return -1;
}  

int 
simpad_get_power(void)
{
	return 1;
}






int 
corgi_set_level(int level)
{
	int fd, val, len, res = -1;
	char buf[20];
	
	if ((fd = open(CORGI_FL, O_WRONLY|O_NONBLOCK)) >= 0)
	{
		val = (level == 1) ? 1 : level * ( 17.0 / 255.0 );
		len = snprintf(buf, 20, "0x%x\n", val);
		res = (write (fd, buf, len) >= 0);
		close (fd);
    }
	return ((res < 0) ? -1 : level);
}

int 
corgi_set_power(int level)
{
	return 1;
}

int 
corgi_get_level(void)
{
	int val, res;
	FILE *fd;
	
	if ((fd = fopen(CORGI_FL, O_RDONLY)) != NULL)
	{
		res = fscanf(fd, "0x%x", &val);
		fclose (fd);
		if (res) 
			return ((val * 255) / 17);
    }
	return (-1);
}

int 
corgi_get_power(void)
{
	return 1;
}






int 
zaurus_set_level(int level)
{
	int fd, val, res = -1;
	
	if ((fd = open(ZAURUS_FL, O_WRONLY)) >= 0)
	{
		val = (level * 4 + 127) / 255;
		if (level && !val)
			val = 1;
            res = ioctl(fd, FL_IOCTL_STEP_CONTRAST, val);
		close (fd);
    }
	return (res != 0) ? -1 : level;
}

int 
zaurus_set_power(int level)
{
	return 1;
}

int 
zaurus_get_level(void)
{
/*	int fd, val;
	
	if ((fd = open(ZAURUS_FL, O_WRONLY )) >= 0 )
	{
		res = ioctl(fd, FL_IOCTL_GET_STEP_CONTRAST, &val);
		close (fd);
    }
	return (res != 0) ? -1 : ((val*255-127) / 4);
*/
	return 25;	
}

int 
zaurus_get_power(void)
{
	return 1;
}






int 
ipaq_set_level(int level)
{
	int light_fd;
	struct h3600_ts_backlight bl;
	float factor;

	switch (IpaqModel) {
		case H36XX:
			factor=1;
			break;
		case H38XX:
			factor = 64/255;
			break;
		case H5XXX:
			factor=1;
			break;
		case H19XX:
			factor=1;
			break;
		case H22XX:
			factor=1;
			break;
		default:
			factor=1;
		break;
	}


	light_fd = open (TS_DEV, O_RDONLY);
	if (light_fd > 0) {
  		bl.brightness = (unsigned char)level*factor;
  		bl.power = (bl.brightness > 0) ? FLITE_PWR_ON : FLITE_PWR_OFF;
 		if (ioctl (light_fd, TS_SET_BACKLIGHT, &bl) != 0)
		{
			close(light_fd);
    		return -1;
		}
  		else
		{
			close(light_fd);
    		return level/factor;
		}
	}
	return -1;
}

int 
ipaq_set_power(int power_on)
{
	int light_fd;
	struct h3600_ts_backlight bl;

	light_fd = open (TS_DEV, O_RDONLY);
	if (light_fd > 0) {

  		bl.brightness = (unsigned char)ipaq_get_level();

		if (power_on == 1)
		{
  			bl.power = FLITE_PWR_ON;
		}
		else
		{
  			bl.power = FLITE_PWR_OFF;
		}

 		if (ioctl (light_fd, TS_SET_BACKLIGHT, &bl) != 0)
		{
			close(light_fd);
    		return -1;
		}
  		else
		{
			close(light_fd);
    		return bl.power;
		}
	}
	return -1;
}

int 
ipaq_get_level(void)
{
	int light_fd;
	float factor;
	struct h3600_ts_backlight bl;

	switch (IpaqModel) {
		case H36XX:
			factor=1;
			break;
		case H38XX:
			factor = 64/255;
			break;
		case H5XXX:
			factor=1;
			break;
		case H19XX:
			factor=1;
			break;
		case H22XX:
			factor=1;
			break;
		default:
			factor=1;
		break;
	}

	light_fd = open (TS_DEV, O_RDONLY);
	if (light_fd > 0) {
 		if (ioctl (light_fd, TS_GET_BACKLIGHT, &bl) != 0)
		{
			close(light_fd);
    		return -1;
		}
  		else
		{
			close(light_fd);
    		return (int)bl.brightness/factor;
		}
	}
	return -1;
}

int 
ipaq_get_power(void)
{
	int light_fd;
	struct h3600_ts_backlight bl;

	light_fd = open (TS_DEV, O_RDONLY);
	if (light_fd > 0) {
 		if (ioctl (light_fd, TS_GET_BACKLIGHT, &bl) != 0)
		{
			close(light_fd);
    		return -1;
		}
  		else
		{
			close(light_fd);
    		if (bl.power == FLITE_PWR_ON)
			{
				return 1;
			}
			else
			{
				return 0;
			}
		}
	}
	return -1;
}



int 
integral_set_level(int level)
{
  FILE *f_light;
  
  f_light = fopen(PROC_LIGHT,"w");
  if (f_light != NULL)
  {
  	fprintf(f_light,"%i\n", level);
  	fclose(f_light);
	return level;
  }
  else
	  return -1;
}

int 
integral_set_power(int level)
{
	return 1;
}

int 
integral_get_level(void)
{
  FILE *f_light;
  int level;
  
  f_light = fopen(PROC_LIGHT,"r");
  if (f_light != NULL)
  {
  	fscanf(f_light,"%i", &level);
  	fclose(f_light);
	return level;
  }
  return -1;
}  
  
int 
integral_get_power(void)
{
	return 1;
}

int
get_ipaq_model (void)
{
  int fd;
  char buf[16];

  fd = open ("/proc/hal/model", O_RDONLY);
  if (fd <= 0)
    return -1;
  if ( read( fd, buf, 16) < 4)
    return -1;
  close(fd);

  if ((strncmp (buf, "36", 2) == 0) || (strncmp (buf, "37", 2) == 0)) {
    /*g_print ("model is H36xx\n");*/
    IpaqModel = H36XX;
  } else if ((strncmp (buf, "38", 2) == 0) || (strncmp (buf, "39", 2) == 0)) {
    /*g_print ("model is H38xx\n");*/
    IpaqModel = H38XX;
  } else if ((strncmp (buf, "54", 2) == 0) || (strncmp (buf, "55", 2) == 0)) {
    /*g_print ("model is H5xxx\n");*/
    IpaqModel = H5XXX;
  } else
    IpaqModel = -1;

return IpaqModel;
}



/*
 Set the backlight level: 0-255
  */

int 
set_backlight_level (int level)
{
	if (level < 0)
		level = 0;
	if (level > 255)
		level = 255;
	switch (platform)
	{
	case P_IPAQ:
		return ipaq_set_level(level);
	break;
	case P_ZAURUS:
		return zaurus_set_level(level);
	break;
	case P_CORGI:
		return corgi_set_level(level);
	break;
	case P_INTEGRAL:
		return integral_set_level(level);
	break;
	case P_SIMPAD_NEW:
		return simpad_new_set_level(level);
	break;
	case P_SIMPAD:
		return simpad_set_level(level);
	break;
	case P_GENERIC:
		level = level/255*32;
		return generic_set_level(level);
	break;
	case P_SYSCLASS:
	case P_SYSCLASS_770:
		return sysclass_set_level(level);
	break;	
	default:
		return 0;
	break;
	}
}


/*
 Set the backlight power: 1=on, 0=off
  */

int 
set_backlight_power (int power_on)
{
	switch (platform)
	{
	case P_IPAQ:
		return ipaq_set_power(power_on);
	break;
	case P_ZAURUS:
		return zaurus_set_power(power_on);
	break;
	case P_CORGI:
		return corgi_set_power(power_on);
	break;
	case P_INTEGRAL:
		return integral_set_power(power_on);
	break;
	case P_SIMPAD_NEW:
		return simpad_new_set_power(power_on);
	break;
	case P_SIMPAD:
		return simpad_set_power(power_on);
	break;
	case P_GENERIC:
		return generic_set_power(power_on);
	break;
	case P_SYSCLASS:
	case P_SYSCLASS_770:
		return sysclass_set_power(power_on);
	break;	
	default:
		return 0;
	break;
	}
}


/*
 Get the backlight level: 0-255
  */

int
get_backlight_level (void)
{
	switch (platform)
	{
	case P_IPAQ:
		return ipaq_get_level();
	break;
	case P_ZAURUS:
		return zaurus_get_level();
	break;
	case P_CORGI:
		return corgi_get_level();
	break;
	case P_INTEGRAL:
		return integral_get_level();
	break;
	case P_SIMPAD_NEW:
		return simpad_new_get_level();
	break;
	case P_SIMPAD:
		return simpad_get_level();
	break;
	case P_GENERIC:
		return generic_get_level();
	break;
	case P_SYSCLASS:
	case P_SYSCLASS_770:
		return sysclass_get_level();
	break;	
	default:
		return 0;
	break;
	}
  return 0;
}


/*
 Get the backlight power: 0=on, 1=off
  */

int 
get_backlight_power (void)
{
	switch (platform)
	{
	case P_IPAQ:
		return ipaq_get_power();
	break;
	case P_ZAURUS:
		return zaurus_get_power();
	break;
	case P_CORGI:
		return corgi_get_power();
	break;
	case P_INTEGRAL:
		return integral_get_power();
	break;
	case P_SIMPAD_NEW:
		return simpad_new_get_power();
	break;
	case P_SIMPAD:
		return simpad_get_power();
	break;
	case P_GENERIC:
		return generic_get_power();
	break;
	case P_SYSCLASS:
	case P_SYSCLASS_770:
		return sysclass_get_power();
	break;	
	default:
		return 0;
	break;
	}
}




int main(int argc, char ** argv )
{
	int retval;

	if ( argc > 1 && argv[1][0] == '-' )
		usage();


	platform = detect_platform();

	if (platform == P_IPAQ)
		if (get_ipaq_model() == -1) {
		exit (1);
		}



	switch (argc) {
	case 1:   /* No arguments */
		printf("%s %d\n", (get_backlight_power() == FLITE_PWR_OFF ? "off" : "on"), get_backlight_level() );
		break;
	case 2:   /* One argument */
		if ( !strcasecmp(argv[1],"off") ) {
			if (debug) printf("Turning off\n");
			set_backlight_power(FLITE_PWR_OFF);   /* Don't adjust the current brightness */
		}
		else if ( !strcasecmp(argv[1],"on")) {
			if (debug) printf("Turning on\n");
			set_backlight_power(FLITE_PWR_ON);
		}
		else if ( !strcasecmp(argv[1], "toggle")) {
			int power = (get_backlight_power() == FLITE_PWR_OFF) ? FLITE_PWR_ON : FLITE_PWR_OFF;
			set_backlight_power(power);
			if (debug) printf("Turning %s\n", (get_backlight_power() == FLITE_PWR_OFF) ? "off" : "on");
		}
		else {
			set_backlight_power(FLITE_PWR_ON);
			set_backlight_level(atoi(argv[1]));
			if (debug) printf("Turning to %d\n", get_backlight_level());
		}
		break;
	case 4:   /* Three arguments */
		/* int mode = (unsigned char)atoi(argv[1]); */
		set_backlight_power(atoi(argv[2]));
		set_backlight_level(atoi(argv[3]));
		break;
	default:
		usage();
		break;
	}

	return 0;
}


